;;; howm-view.el --- Wiki-like keyword jump between files
;;; Copyright (c) 2002, 2003, 2004
;;;   by HIRAOKA Kazuyuki <khi@users.sourceforge.jp>
;;; $Id: howm-view.el,v 1.174 2004/11/26 12:08:53 hira Exp $
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 1, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; The GNU General Public License is available by anonymouse ftp from
;;; prep.ai.mit.edu in pub/gnu/COPYING.  Alternately, you can write to
;;; the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
;;; USA.
;;--------------------------------------------------------------------

(require 'howm-common)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; variables

;; customize
(defvar howm-view-summary-persistent t
  "If non-nil, keep summary buffer on howm-view-summary-open.
If it is a function, the evaluated value is used instead of itself.")
(defvar howm-view-contents-persistent t
  "If non-nil, keep contents buffer on howm-view-contents-open.
If it is a function, the evaluated value is used instead of itself.")
(defvar howm-view-contents-limit nil
  "Max length for howm-view-contents. Nil means no limit.")
; (defvar howm-view-contents-limit (* 80 200))
(defvar howm-view-summary-keep-cursor nil)
(defvar howm-view-summary-mode-hook nil)
(defvar howm-view-contents-mode-hook nil)
(defvar howm-view-open-hook nil)
(defvar howm-view-before-open-hook nil)
(defvar howm-view-hilit-face 'howm-view-hilit-face)
(defface howm-view-hilit-face '((t (:foreground "red")))
  "Face for matched word.")
(defvar howm-view-summary-sep "|")
(defvar howm-view-summary-format
  (let* ((path (format-time-string howm-file-name-format))
         (width (length (file-name-nondirectory path))))
    (concat "%-" (format "%s" width) "s " howm-view-summary-sep " ")))
;;     (concat "%-" (format "%s" width) "s | ")))
(defvar howm-view-header-format
  "\n==========================>>> %s\n"
  "Format string of header for howm-view-contents.
%s is replaced with file name. See `format'.")
(defvar howm-view-header-regexp "^==========================>>> .*$")
(defvar howm-view-open-recenter howm-view-search-recenter)
(defvar howm-view-title-header "=")
;; howm-view-title-regexp is assumed to have a form "^xxxxxxx$"
(defvar howm-view-title-regexp (format "^%s\\( +\\(.*\\)\\|\\)$"
                                     (regexp-quote howm-view-title-header)))
(defvar howm-view-title-regexp-pos 2)
(defvar howm-view-title-regexp-grep (format "^%s +"
                                     (regexp-quote howm-view-title-header)))
(defun howm-view-title-regexp-grep ()
  (if howm-view-use-grep
      howm-view-title-regexp-grep
    howm-view-title-regexp))

(defvar howm-view-use-grep nil)
(defvar howm-view-grep-command "egrep")
(defvar howm-view-fgrep-command "fgrep")
(defvar howm-view-grep-option "-Hnr")
(defvar howm-view-grep-ignore-case-option "-i")
(defvar howm-view-grep-expr-option "-e")
;; (defvar howm-view-grep-file-stdin-option "-f -")
(defvar howm-view-grep-file-stdin-option nil)

(defvar howm-view-sort-methods
  '(("random" . howm-view-sort-by-random)
    ("name" . howm-view-sort-by-name)
    ("name-match" . howm-view-sort-by-name-match)
    ("summary" . howm-view-sort-by-summary)
    ("summary-match" . howm-view-sort-by-summary-match)
;     ("atime" . howm-view-sort-by-atime) ;; nonsense
;     ("ctime" . howm-view-sort-by-ctime) ;; needless
    ("mtime" . howm-view-sort-by-mtime)
    ("date" . howm-view-sort-by-reverse-date)
    ("reminder" . howm-view-sort-by-reminder)
    ("reverse" . howm-view-sort-reverse)))

(defvar howm-view-filter-methods
  '(("name" . howm-view-filter-by-name)
    ("summary" . howm-view-filter-by-summary)
    ("mtime" . howm-view-filter-by-mtime)
;     ("ctime" . howm-view-filter-by-ctime) ;; needless
    ("date" . howm-view-filter-by-date)
    ("reminder" . howm-view-filter-by-reminder)
    ("contents" . howm-view-filter-by-contents)
    ("Region" . howm-view-filter-by-region)
    ("Around" . howm-view-filter-by-around)
;     ("uniq" . howm-view-filter-uniq))
  ))

;; referred only when howm-view-use-grep is nil
(defvar howm-view-watch-modified-buffer t)

(defvar howm-view-update-search-ring nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; item

;; item = (filename summary place offset home)
(defun howm-view-make-item (filename &optional summary place offset home
                                     privilege)
  (list filename summary place offset home privilege))
(defun howm-view-item-filename      (item) (howm-view-item-nth 0 item ""))  
(defun howm-view-item-summary       (item) (howm-view-item-nth 1 item ""))  
(defun howm-view-item-place         (item) (howm-view-item-nth 2 item nil)) 
(defun howm-view-item-offset        (item) (howm-view-item-nth 3 item nil)) 
(defun howm-view-item-home          (item) (howm-view-item-nth 4 item nil)) 
(defun howm-view-item-privilege     (item) (howm-view-item-nth 5 item nil)) 
(defun howm-view-item-nth (n item default)
  (or (nth n item) default))
(defun howm-view-item-set-summary (item val)
  (setf (nth 1 item) val))
(defun howm-view-item-set-offset (item val)
  (setf (nth 3 item) val))
(defun howm-view-item-set-home (item val)
  (setf (nth 4 item) val))
(defun howm-view-item-set-privilege (item val)
  (setf (nth 5 item) val))

(defun howm-view-item-basename (item &optional nonempty)
  (let* ((f (howm-view-item-filename item))
         (b (file-name-nondirectory f)))
    (if (and (string= b "") nonempty)
        f
      b)))
(defun howm-view-item-exclude-p (name-checker item)
  (and (not (howm-view-item-privilege item))
       (funcall name-checker (howm-view-item-filename item))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; riffle

;;; controller

(defvar howm-view-controller
  (riffle-make-controller
   '(
;;      (section . howm-view-item-filename)
     (home . howm-view-item-home)
     (summary-item  . howm-view-summary-item)
     (contents-item   . howm-view-contents-item)
     (summary-mode  . howm-view-summary-mode)
     (contents-mode . howm-view-contents-mode)
     (summary-name-format  . (lambda () howm-view-summary-name))
     (contents-name-format . (lambda () howm-view-contents-name))
     (post-update . (lambda (item)
                      (message "View: %s" (howm-view-item-filename item))))
     )))

;;; aliases

;; Only howm-view.el should call riffle-xxx.
;; Define alias if it is used in howm-xxx besides howm-view.el.
(defalias 'howm-view-name          #'riffle-name)          
(defalias 'howm-view-item-list     #'riffle-item-list)     
(defalias 'howm-view-line-number   #'riffle-line-number)   
(defalias 'howm-view-summary-check #'riffle-summary-check) 
(defalias 'howm-view-persistent-p  #'riffle-persistent-p)  
(defalias 'howm-view-kill-buffer   #'riffle-kill-buffer)   
(defalias 'howm-view-set-place     #'riffle-set-place)     
(defalias 'howm-view-summary-to-contents 'riffle-summary-to-contents)
(defalias 'howm-view-restore-window-configuration #'riffle-restore-window-configuration)

;; for howmoney.el
;; http://howm.sourceforge.jp/cgi-bin/hiki/hiki.cgi?howmoney
(defun howm-view-get-buffer (name-format &optional name new)
  (let ((riffle-controller howm-view-controller)) ;; cheat
    (riffle-get-buffer name-format name new)))
(defun howm-view-summary-buffer (&optional new)
  (let ((riffle-controller howm-view-controller)) ;; cheat
    (riffle-summary-buffer new)))
(defalias 'howm-view-summary-show 'riffle-summary-show)

;;; variables

(defvar howm-view-summary-name "*howmS*"
  "Format string of buffer name for summary.
%s is replaced with searched string. See `format'.")
(defvar howm-view-contents-name "*howmC*"
  "Format string of buffer name for contents.
%s is replaced with searched string. See `format'.")
;; (defvar howm-view-summary-name "*howmS:%s*"
;;   "Format string of buffer name for summary.
;; %s is replaced with searched string. See `format'.")
;; (defvar howm-view-contents-name "*howmC:%s*"
;;   "Format string of buffer name for contents.
;; %s is replaced with searched string. See `format'.")

(defvar howm-view-font-lock-silent t
  "Inhibit font-lock-verbose if non-nil.")
(defvar howm-view-summary-font-lock-keywords nil)
(defvar howm-view-contents-font-lock-keywords nil)

(defvar *howm-view-font-lock-keywrods* nil)
(defun howm-view-font-lock-keywords () *howm-view-font-lock-keywrods*)

;;; modes

(define-derived-mode howm-view-summary-mode riffle-summary-mode "HowmS"
  "memo viewer (summary mode)
key	binding
---	-------
\\[howm-view-summary-open]	Open file
\\[next-line]	Next item
\\[previous-line]	Previous item
\\[riffle-pop-or-scroll-other-window]	Pop and scroll contents
\\[scroll-other-window-down]	Scroll contents
\\[riffle-scroll-other-window]	Scroll contents one line
\\[riffle-scroll-other-window-down]	Scroll contents one line
\\[riffle-summary-to-contents]	Concatenate all contents
\\[howm-view-filter-uniq]	Remove duplication of same file
\\[howm-view-summary-shell-command]	Execute command in inferior shell

\\[delete-other-windows]	Delete contents window
\\[riffle-pop-window]	Pop contents window
\\[riffle-toggle-window]	Toggle contents window
\\[howm-list-title]	Show Title

\\[howm-view-filter]	Filter (by date, contents, etc.)
\\[howm-view-filter-by-contents]	Search (= filter by contents)
\\[howm-view-sort]	Sort (by date, summary line, etc.)
\\[howm-view-sort-reverse]	Reverse order
\\[howm-view-dired]	Invoke Dired-X
\\[describe-mode]	This help
\\[riffle-kill-buffer]	Quit
"
  (make-local-variable 'font-lock-keywords)
  (make-local-variable 'font-lock-keywords-only)
  (make-local-variable 'font-lock-keywords-case-fold-search)
  (cheat-font-lock-mode howm-view-font-lock-silent)
  (when (howm-view-font-lock-keywords)
    (cheat-font-lock-merge-keywords (howm-view-font-lock-keywords)
                                    howm-view-summary-font-lock-keywords))
  (setq font-lock-keywords-only t)
  (setq font-lock-keywords-case-fold-search t)
  ;;     (setq font-lock-keywords-case-fold-search
  ;;           howm-view-grep-ignore-case-option)
  (font-lock-fontify-block)
  )

(define-derived-mode howm-view-contents-mode riffle-contents-mode "HowmC"
  "memo viewer (contents mode)
key	binding
---	-------
\\[howm-view-contents-open]	Open file
\\[next-line]	Next line
\\[previous-line]	Previous line
\\[scroll-up]	Scroll up
\\[scroll-down]	Scroll down
\\[riffle-scroll-up]	Scroll one line up
\\[riffle-scroll-down]	Scroll one line down
\\[riffle-contents-to-summary]	Summary
\\[riffle-contents-goto-next-item]	Next item
\\[riffle-contents-goto-previous-item]	Previous item

\\[howm-view-filter]	Filter (by date, contents, etc.)
\\[howm-view-filter-by-contents]	Search (= filter by contents)
\\[howm-view-sort]	Sort
\\[howm-view-sort-reverse]	Reverse order
\\[howm-view-dired]	Invoke Dired-X
\\[describe-mode]	This help
\\[riffle-kill-buffer]	Quit
"
;   (kill-all-local-variables)
  (make-local-variable 'font-lock-keywords)
  (make-local-variable 'font-lock-keywords-only)
  (make-local-variable 'font-lock-keywords-case-fold-search)
  (cheat-font-lock-mode howm-view-font-lock-silent)
  (let ((ck `((,howm-view-header-regexp (0 howm-view-hilit-face))))
        ;; extremely dirty!! [2003/10/06 21:08]
        (sk (or (with-current-buffer (riffle-summary-buffer)
                  font-lock-keywords)
                *howm-view-font-lock-keywrods*)))
    (cheat-font-lock-merge-keywords howm-view-contents-font-lock-keywords
                                    sk ck)
    (setq font-lock-keywords-only t)
    (setq font-lock-keywords-case-fold-search
          howm-view-grep-ignore-case-option)
;;    (font-lock-fontify-block)
;;     (run-hooks 'howm-view-contents-mode-hook)
    (font-lock-fontify-block)
    ))

;;; keymaps

;; (defvar howm-view-summary-mode-map nil)
;; (defvar howm-view-contents-mode-map nil)

(defun howm-view-define-common-key (keymap)
  (let ((m keymap))
;;     (define-key m "?" 'howm-view-help)
    (define-key m "f" 'howm-view-filter)
    (define-key m "G" 'howm-view-filter-by-contents)
    (define-key m "S" 'howm-view-sort)
    (define-key m "R" 'howm-view-sort-reverse)
    (define-key m "q" 'howm-view-kill-buffer)
    (define-key m "X" 'howm-view-dired)
    ))

(let ((m howm-view-summary-mode-map))
  (define-key m "\C-m" 'howm-view-summary-open)
  (define-key m "\C-j" 'howm-view-summary-open)
  (define-key m "u" 'howm-view-filter-uniq)
  (define-key m "!" 'howm-view-summary-shell-command)
  (define-key m "T" 'howm-list-title) ;; defined in other file. dirty!
  ;;     (define-key m howm-reminder-quick-check-key 'howm-reminder-quick-check)
  ;;     (define-key m ";" 'howm-view-invoke-action-lock)
  (define-key m "\C-i" 'howm-view-summary-next-section)
  (define-key m "\M-\C-i" 'howm-view-summary-previous-section)
  (define-key m [tab] 'howm-view-summary-next-section)
  (define-key m [(meta tab)] 'howm-view-summary-previous-section)
  (howm-view-define-common-key m))

(let ((m howm-view-contents-mode-map))
  (define-key m "\C-m" 'howm-view-contents-open)
  (define-key m "\C-j" 'howm-view-contents-open)
  (howm-view-define-common-key m))

;;; summary

(defun howm-view-summary (&optional name item-list)
  (riffle-summary name item-list howm-view-controller))

(defun howm-view-summary-open (&optional reverse-delete-p)
  (interactive "P")
  (when (not (and howm-view-summary-keep-cursor
                  (get-buffer-window (riffle-contents-buffer))))
    (riffle-summary-check t))
  (let* ((p (riffle-persistent-p howm-view-summary-persistent))
         (persistent (if reverse-delete-p
                         (not p)
                       p)))
    (howm-record-view-window-configuration)
    (howm-view-summary-open-sub (not persistent))))

(defun howm-view-summary-open-sub (&optional kill)
  (interactive "P")
  (let ((b (riffle-contents-buffer))
        (looking-at-str (buffer-substring-no-properties (point)
                                                        (line-end-position))))
    (riffle-pop-to-buffer b howm-view-summary-window-size)
    (let ((howm-view-open-hook nil)) ;; Don't execute it in contents-open.
      (howm-view-contents-open-sub kill))
    (end-of-line)
    (or (search-backward looking-at-str (line-beginning-position) t)
        (beginning-of-line))
    (run-hooks 'howm-view-open-hook)))

(defun howm-view-summary-item (item)
  (let* ((f (howm-view-item-basename item t))
         (h (format howm-view-summary-format f)))
    (concat h (howm-view-item-summary item))))

(defun howm-view-summary-next-section (&optional n)
  (interactive "P")
  (setq n (or n 1))
  (let ((i (abs n))
        (step (if (>= n 0) 1 -1)))
    (while (and (> i 0)
                (howm-view-summary-next-section-sub step))
      (setq i (1- i)))))
(defun howm-view-summary-previous-section (&optional n)
  (interactive "P")
  (setq n (or n 1))
  (howm-view-summary-next-section (- n)))
(defun howm-view-summary-next-section-sub (step)
  (let ((orig (howm-view-item-filename (riffle-summary-current-item))))
;;   (let ((orig (riffle-controller 'section
;;                                  (riffle-summary-current-item))))
    (while (and (string= orig
                         (howm-view-item-filename (riffle-summary-current-item)))
;;                          (riffle-controller 'section
;;                                             (riffle-summary-current-item)))
                (= (forward-line step) 0))
      ;; no body
      )))

;;; contents

(defun howm-view-contents-open (&optional reverse-delete-p)
  (interactive "P")
  (let* ((p (riffle-persistent-p howm-view-contents-persistent))
         (persistent (if reverse-delete-p
                         (not p)
                       p)))
    (howm-record-view-window-configuration)
    (howm-view-contents-open-sub (not persistent))))

(defvar *howm-view-item-privilege* nil) ;; dirty

(defun howm-view-contents-open-sub (&optional kill)
  (let* ((n (riffle-contents-item-number (point)))
         (item (nth n (riffle-item-list)))
         (file (howm-view-item-filename item))
         (offset (howm-view-item-offset item))
         (pos (- (point) offset))
         (viewer (howm-view-external-viewer file)))
    (when kill
      (riffle-kill-buffer))
    (when (howm-view-item-privilege item)
      (riffle-restore-window-configuration)) ;; force without mode check
    (setq *howm-view-item-privilege* (howm-view-item-privilege item)) ;; dirty
    (run-hooks 'howm-view-before-open-hook)
    (if viewer
        (howm-view-call-external-viewer viewer file)
      (progn
        (find-file file)
        (howm-view-set-mark-command)
        (when (or (< pos (point-min)) (< (point-max) pos))
          (widen))
        (goto-char pos)
        (recenter howm-view-open-recenter)))
    (run-hooks 'howm-view-open-hook)))

(defun howm-view-insert-file/buffer-contents (file)
  (let ((b (get-file-buffer file)))
    (if (and b
             howm-view-watch-modified-buffer
             (not howm-view-use-grep))
        (insert (with-current-buffer b
                  (save-restriction
                    (widen)
                    (buffer-substring-no-properties (point-min) (point-max)))))
      (insert-file-contents file))))

(defvar howm-view-previous-section-name nil "For internal use") 
(defvar howm-view-previous-section-beg nil "For internal use") 
(defvar howm-view-previous-section-end nil "For internal use") 

(defun howm-view-contents-item (item)
  (when (howm-buffer-empty-p)
    (setq howm-view-previous-section-name ""
          howm-view-previous-section-beg nil
          howm-view-previous-section-end nil))
  (let* ((fname (howm-view-item-filename item))
         (place (howm-view-item-place item))
         (done-p (if place
                     (and (string= fname howm-view-previous-section-name) ;; dirty!
                          (<= howm-view-previous-section-beg place)
                          (< place howm-view-previous-section-end))
                   (string= fname howm-view-previous-section-name))))
    (if done-p
        ""
      (let* ((header (format howm-view-header-format
                             (howm-abbreviate-file-name fname)))
             (header-length (howm-view-string-point-count header))
             (viewer (howm-view-external-viewer fname)))
        (concat header
                (howm-view-contents-item-sub item fname place header viewer
                                             (+ (point) header-length)))))))

(defvar howm-view-string-point-count-strict nil)
(defun howm-view-string-point-count (str)
  "Count points of string STR.
Namely, it is the difference between start position and end position
of STR if STR is inserted to a buffer.
It looks to be simply equal to (length STR) on emacs-21.1.1.
But I'm not sure for multi-byte characters on other versions of emacsen."
  (if howm-view-string-point-count-strict
      (with-temp-buffer
        (insert str)
        (- (point) (point-min)))
    ;; I assume (length (buffer-substring-no-properties START END))
    ;; is equal to (abs (- START END))). Is it correct?
    ;; (cf.) snap://Info-mode/elisp#Positions
    (length str)))

(defun howm-view-contents-item-sub (item fname place header viewer c)
  (with-temp-buffer
    (let (b e h)
      (if viewer
          (howm-view-contents-indicator viewer fname)
        (howm-view-insert-file/buffer-contents fname))
      (if place
          (progn
            (riffle-set-place place)
            (setq h (point))
            (let ((r (howm-view-contents-region fname)))
              (setq b (car r)
                    e (second r))))
        (setq b (point-min)
              e (point-max)
              h b))
      (howm-view-item-set-offset item (- c b))
      (howm-view-item-set-home item (+ c (- b) h))
      (setq howm-view-previous-section-name fname ;; dirty!
            howm-view-previous-section-beg (riffle-get-place b)
            howm-view-previous-section-end (riffle-get-place e))
      (buffer-substring-no-properties b e))))

(defvar howm-view-preview-narrow t)
(defun howm-view-contents-region (filename)
  (when filename
    (howm-set-configuration-for-file-name filename))
  (if (or howm-view-preview-narrow
          (not (riffle-preview-p)))
      (howm-view-paragraph-region)
    (list (point-min) (point-max))))

(defun howm-view-contents-indicator (viewer fname)
  (insert "%%% "
          (cond ((stringp viewer) (format viewer fname))
                ((functionp viewer) (format "%s" viewer))
                (t "Unknown viewer"))
          " %%%"))

(defun howm-view-paragraph-region (&optional include-following-blank-p)
  (let ((b (save-excursion
             (end-of-line)
             (re-search-backward howm-view-title-regexp
                                 nil 'to-limit)
             (line-beginning-position)))
        (e (save-excursion
             (end-of-line)
             (let ((found (re-search-forward howm-view-title-regexp
                                             nil 'to-limit)))
               (if include-following-blank-p
                   (if found (match-beginning 0) (point-max))
                 (progn
                   (if found
                       (forward-line -1)
                     (goto-char (point-max)))
                   (end-of-line)
                   (while (and (looking-at "^$")
                               (= (forward-line -1) 0)) ;; successful
                     nil) ;; dummy
                   (end-of-line)
                   (point)))))))
    (list b e)))

(defun howm-view-set-mark-command ()
  (set-mark-command nil)
  (howm-deactivate-mark))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; misc.

(defun howm-view-file-list ()
  (remove-duplicates (mapcar #'howm-view-item-filename (howm-view-item-list))
                     :test #'string=))

(defun howm-view-xtime (file x)
  (let* ((a (file-attributes file))
         (n (cdr (assoc x '((a . 4) (m . 5) (c . 6)))))
         (ti (nth n a)))
    (howm-view-time-to-string ti)))

(defun howm-view-time-to-string (ti)
  (format-time-string "%Y%m%d-%H%M%S" ti))

(defun howm-view-string> (a b)
  (string< b a))

(defun howm-view-string<= (a b)
  (not (string< b a)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; dir

(defun howm-view-directory (dir &optional recursive-p)
    (howm-view-summary "" (howm-view-directory-items dir recursive-p))
    ;; sloppy!
    ;; (for old ruby-mode.el which sets global value of font-lock-keywords)
    (setq font-lock-keywords nil)
    )

(defun howm-view-directory-items (dir &optional recursive-p)
  (let* ((files (if recursive-p
                    (howm-view-files dir)
                  (directory-files dir t)))
         (summary ""))
    (mapcar (lambda (f)
              (howm-view-make-item f summary))
            files)))

;; list files recursively
(defun howm-view-files (dir &optional exclusion-checker)
  (when (null exclusion-checker)
    (setq exclusion-checker (lambda (x) nil)))
  (cond ((file-directory-p dir) (howm-view-files-sub dir exclusion-checker))
        ((file-exists-p dir) (list dir))
        (t nil)))

(defun howm-view-files-sub (dir exclusion-checker)
  (setq dir (expand-file-name dir))
  (mapcan (lambda (f)
            (cond
             ((funcall exclusion-checker f) nil)
             ((file-directory-p f) (if (howm-subdirectory-p f dir t)
                                       (howm-view-files f exclusion-checker)
                                     nil)) ;; exclude "." & ".."
             ((file-regular-p f) (list f))
             (t nil)))
          (directory-files dir t)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; filter

(defun howm-view-filter (&optional remove-p)
  (interactive "P")
  (let* ((table howm-view-filter-methods)
         (command (completing-read (if remove-p
                                       "(Reject) filter by: "
                                     "filter by: ")
                                   table nil t)))
    (call-interactively (cdr (assoc command table)))))

(defvar howm-view-filter-uniq-prev "")
(defun howm-view-filter-uniq ()
  (interactive)
  (howm-view-filter-general
   (lambda (item)
     (let ((f (howm-view-item-filename item)))
       (prog1
           (not (string= f howm-view-filter-uniq-prev))
         (setq howm-view-filter-uniq-prev f))))))

(defun howm-view-filter-by-name (&optional remove-p regexp)
  (interactive "P")
  (howm-view-filter-by-name/summary 'howm-view-item-basename
                                    remove-p regexp))

(defun howm-view-filter-by-summary (&optional remove-p regexp)
  (interactive "P")
  (howm-view-filter-by-name/summary 'howm-view-item-summary
                                    remove-p regexp))

(defun howm-view-filter-by-name/summary (accessor remove-p regexp)
  (let* ((r (or regexp (howm-view-filter-read-from-minibuffer "Regexp: "
                                                              remove-p))))
    (howm-view-filter-general (lambda (item)
                                (string-match r (funcall accessor item)))
                              remove-p)))

(defun howm-view-filter-by-date (&optional remove-p)
  (interactive "P")
  (let* ((r (howm-view-ask-time-range remove-p))
         (from (car r))
         (to (second r))
         (form (howm-view-file-name-format))
         (fts (mapcar (lambda (x)
                        (file-name-nondirectory (format-time-string form x)))
                      (list from to)))
         (fs (car fts))
         (ts (second fts)))
    (howm-view-filter-general
     (lambda (item)
       (let ((cs (howm-view-item-basename item)))
         (and (howm-view-string<= fs cs) (howm-view-string<= cs ts))))
     remove-p)
    ))

(defun howm-view-filter-by-reminder (&optional remove-p)
  (interactive "P")
  (let* ((r (howm-view-ask-time-range remove-p))
         (from (car r))
         (to (second r))
         (from-str (format-time-string howm-date-format from))
         (to-str (format-time-string howm-date-format to))
         (reg (howm-reminder-regexp howm-reminder-types)))
    (howm-view-filter-general
     (lambda (item)
       (let ((s (howm-view-item-summary item)))
         (and (string-match reg s)
              (let* ((x (match-string-no-properties 0 s)) ;; [2004-02-07]@
                     (d (and (string-match howm-date-regexp x)
                             (match-string-no-properties 0 x)))) ;; [2004-02-07]
                (and (howm-view-string<= from-str d)
                     (howm-view-string<= d to-str))))))
     remove-p)
    ))

(defun howm-view-file-name-format ()
  howm-file-name-format) ;; defined in howm-common.el

;; needless
; (defun howm-view-filter-by-ctime (&optional remove-p)
;   (interactive "P")
;   (howm-view-filter-by-xtime 'c remove-p))

(defun howm-view-filter-by-mtime (&optional remove-p range)
  (interactive "P")
  (howm-view-filter-by-xtime 'm remove-p range))

(defun howm-view-filter-by-xtime (x remove-p &optional range)
  (let* ((r (or range (howm-view-ask-time-range remove-p)))
         (from (car r))
         (to (second r))
         (fs (howm-view-time-to-string from))
         (ts (howm-view-time-to-string to)))
    (howm-view-filter-general
     (lambda (item)
       (let* ((cs (howm-view-xtime (howm-view-item-filename item) x)))
         (and (howm-view-string<= fs cs) (howm-view-string<= cs ts))))
     remove-p)))

(defun howm-view-ask-time-range (&optional remove-p)
  (let* ((now (current-time))
         (from (howm-view-ask-time "From" now t remove-p))
         (to (howm-view-ask-time "To" from nil remove-p)))
    (list from to)))

(defvar howm-view-min-year 1950)
(defvar howm-view-max-year 2030)
(defun howm-view-ask-time (prompt default &optional from-p remove-p)
  (let* ((z (decode-time default))
         (yd (nth 5 z))
         (md (nth 4 z))
         (dd (nth 3 z)))
    (let (y0 m0 d0 hour0 min0 sec0)
      (if from-p
          (setq y0 howm-view-min-year m0 1 d0 1
                hour0 0 min0 0 sec0 0)
        (setq y0 howm-view-max-year m0 12 d0 'last-day-of-month
              hour0 24 min0 0 sec0 0))
      (let ((y (howm-ask-time-sub prompt "year" yd remove-p)))
        (if (null y)
            (howm-view-encode-time sec0 min0 hour0 d0 m0 y0)
          (let ((m (howm-ask-time-sub prompt "month" md remove-p)))
            (if (null m)
                (howm-view-encode-time sec0 min0 hour0 d0 m0 y)
              (let ((d (or (howm-ask-time-sub prompt "date" dd remove-p) d0)))
                (howm-view-encode-time sec0 min0 hour0 d m y)))))))))

(defun howm-ask-time-sub (prompt ymd default remove-p)
  (let* ((message (format "%s %s (* = no limit) [%d]: " prompt ymd  default))
         (raw (howm-view-filter-read-from-minibuffer message remove-p))
         (n (if (string= raw "")
                default
              (string-to-number raw))))
    (if (= n 0)
        nil
      n)))

(defun howm-view-encode-time (sec min hour d m y)
  (when (eq d 'last-day-of-month)
    (setq m (+ m 1))
    (setq d -1))
  (encode-time sec min hour d m y))

(defun howm-view-filter-by-region (beg end)
  (interactive "r")
  (let ((r (mapcar #'howm-view-line-number (list beg end))))
    (howm-view-filter-by-line-range (car r) (second r))))

(defvar howm-view-filter-by-around-default 10)
(defun howm-view-filter-by-around (&optional distance)
  (interactive "P")
  (let* ((d (or distance howm-view-filter-by-around-default))
         (c (howm-view-line-number)))
    (howm-view-filter-by-line-range (- c d) (+ c d))))

(defun howm-view-filter-by-line-range (beg end)
  (let ((howm-view-filter-by-line-range-beg beg)
        (howm-view-filter-by-line-range-end end))
    (howm-view-filter-general (lambda (item-count)
                                (let* ((line (1+ (second item-count)))
                                       (b howm-view-filter-by-line-range-beg)
                                       (e howm-view-filter-by-line-range-end))
                                  (and (<= b line) (<= line e))))
                              nil t)))
  
(defun howm-view-filter-general (pred &optional remove-p with-index)
  (let* ((item-list (howm-view-item-list))
         (s (if with-index
                (howm-map-with-index #'list item-list)
              item-list))
         (r (if remove-p
                (remove-if pred s)
              (remove-if-not pred s)))
         (filtered (if with-index
                       (mapcar #'car r)
                     r)))
    (howm-view-summary-rebuild filtered)))

(defun howm-view-filter-read-from-minibuffer (message &optional remove-p)
  (read-from-minibuffer (if remove-p
                            (concat "(Reject) " message)
                          message)))

(defun howm-view-summary-rebuild (item-list)
  (let ((name (howm-view-name)))
    (howm-view-summary name item-list)))

(defun howm-view-filter-by-contents (str)
  (interactive "sSearch in result (grep): ")
  (let* ((orig (howm-view-name))
         (name (if (string= orig "")
                   str
                 (format "%s&%s" orig str)))
         (file-list (howm-view-file-list)))
    (howm-write-history str)
    (howm-view-search str file-list name)))

(defun howm-view-list-title (title-regexp)
  (let* ((files (howm-view-file-list))
         (items (howm-view-search-items title-regexp files))
         (name (howm-view-name))
         (kw (howm-view-font-lock-keywords)))
    (let* ((hit-files (remove-duplicates (mapcar #'howm-view-item-filename
                                                 items)
                                         :test #'string=))
           (nohit-files (remove-if (lambda (f) (member f hit-files)) files))
           (nohit-items (mapcar #'howm-view-make-item nohit-files))
           (all-items (append nohit-items items)))
      (let ((*howm-view-font-lock-keywrods* kw)) ;; dirty!
        (howm-view-summary name all-items)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; search

(defun howm-view-search (str file-list &optional
                             name summarizer fixed-p hilit-keywords)
  ;; clean me. str-orig can be string or list of strings.
  (let* ((str-orig str)
         (str-list (if (listp str-orig) str-orig (list str-orig)))
         (str-principal (if (listp str-orig) (car str-orig) str-orig)))
    ;; rename str
    (setq str str-principal)
    (setq name (or name str))
    (when howm-view-update-search-ring
      (isearch-update-ring str (not fixed-p)))
    (let* ((items (howm-view-search-items str-orig
                                          file-list summarizer fixed-p))
           (kw (or hilit-keywords
                   `((,(regexp-opt str-list) . howm-view-hilit-face))))
;;                    `((,(regexp-quote str) . howm-view-hilit-face))))
           )
      (let* ((f (expand-file-name str)))
        (when (file-exists-p f)
          (let ((fi (howm-view-make-item f)))
            (howm-view-item-set-privilege fi t)
            (setq items (cons fi items)))))
      (let ((*howm-view-font-lock-keywrods* kw)) ;; dirty!
        (howm-view-summary name items)))))

(defun howm-view-search-items (str file-list &optional summarizer fixed-p)
  (let ((found (howm-view-grep str file-list fixed-p))
        (summarizer (or summarizer
                        (lambda (file place content)
                          (string-match "^ *\\(.*\\)" content)
                          (match-string-no-properties 1 content)))))
    (mapcar (lambda (z)
              (let ((file (car z))
                    (place (second z))
                    (content (third z)))
                (howm-view-make-item file
                                     (funcall summarizer
                                              file place content)
                                     place)))
            found)))

(defvar *howm-view-force-case-fold-search* nil) ;; dirty!
(defvar howm-view-grep-log-file nil)
;; (defvar howm-view-grep-log-format "%s >>> %s\n")
(defvar howm-view-grep-log-format "> %s | %s")
;; (defvar howm-view-grep-log-format "> %s | %s\n")
;; (defvar howm-view-grep-log-exclude "^%.*%$")

(defun howm-view-grep (str file-list &optional fixed-p)
;;   (when (and howm-view-grep-log-file
;;              (not (string-match howm-view-grep-log-exclude str)))
  (when howm-view-grep-log-file
    (howm-write-log str howm-view-grep-log-format howm-view-grep-log-file))
;;     (save-excursion
;;       (with-current-buffer (find-file-noselect howm-view-grep-log-file)
;;         (goto-char (point-min))
;;         (insert (format howm-view-grep-log-format
;;                         (format-time-string howm-dtime-format (current-time))
;;                         str))
;;         (basic-save-buffer))))
  (when (stringp file-list)
    (setq file-list (list file-list)))
  (let ((grep-func (cond ((eq howm-view-use-grep t) 'howm-view-real-grep)
                         ((null howm-view-use-grep) 'howm-view-fake-grep)
                         ((functionp howm-view-use-grep) howm-view-use-grep)
                         (t (error "No function %s." howm-view-use-grep)))))
    (funcall grep-func
             str file-list fixed-p *howm-view-force-case-fold-search*)))

(defun howm-view-real-grep (str file-list &optional fixed-p force-case-fold)
  "Call grep and parse its result.
'((file line-number line) (file line-number line) ...)
"
  (if (howm-view-grep-multi-p)
      (howm-view-real-grep-multi str file-list fixed-p force-case-fold)
    (howm-view-real-grep-single str file-list fixed-p force-case-fold)))

(defun howm-view-grep-multi-p ()
  howm-view-grep-file-stdin-option)

;; will be obsolete when howm-view-real-grep-multi is stable
(defun howm-view-real-grep-single (str file-list
                                       &optional fixed-p force-case-fold)
  "Call grep and parse its result.
'((file line-number line) (file line-number line) ...)
"
  (let ((grep-command (if fixed-p
                          howm-view-fgrep-command
                        howm-view-grep-command))
        (opt (list howm-view-grep-option))
        (eopt (and howm-view-grep-expr-option
                   (list howm-view-grep-expr-option)))
        (case-fold (or force-case-fold
                       (not (let ((case-fold-search nil))
                              (string-match "[A-Z]" str))))))
    (when (and case-fold howm-view-grep-ignore-case-option)
      (setq opt (cons howm-view-grep-ignore-case-option opt)))
    (with-temp-buffer
      (let* ((fs (mapcar #'expand-file-name file-list))
             (lines (howm-view-call-process grep-command
                                              `(,@opt ,@eopt ,str ,@fs)))
             (parsed (mapcar 'howm-view-grep-parse-line lines)))
        (remove nil parsed)))))

;; experimental [2004-04-11]
(defun howm-view-real-grep-multi (str file-list
                                      &optional fixed-p force-case-fold)
  (let ((grep-command (if fixed-p
                          howm-view-fgrep-command
                        howm-view-grep-command))
        (opt (list howm-view-grep-option))
        (eopt (split-string howm-view-grep-file-stdin-option)))
    (let* ((str-list (cond ((stringp str) (list str))
                           ((listp str) str)
                           (t (error "Wrong type: %s" str))))
           (caps-p (member-if (lambda (s) (howm-capital-p s)) str-list))
;;            (caps-p (let ((case-fold-search nil))
;;                      (member-if (lambda (s) (string-match "[A-Z]" s))
;;                                 str-list)))
           (case-fold (or force-case-fold (not caps-p))))
      (when (and case-fold howm-view-grep-ignore-case-option)
        (setq opt (cons howm-view-grep-ignore-case-option opt)))
      (with-temp-buffer
        (let* ((fs (mapcar #'expand-file-name file-list))
               (pat (apply #'concat
                           (mapcar (lambda (s) (concat s "\n")) str-list)))
               (lines (howm-view-call-process grep-command
                                              `(,@opt ,@eopt ,@fs)
                                              nil pat))
               (parsed (mapcar 'howm-view-grep-parse-line lines)))
          (remove nil parsed))))))

(defun howm-view-fake-grep (str file-list &optional fixed-p force-case-fold)
  (let* ((str-list (if (stringp str) (list str) str))
         (reg (if fixed-p
                  (regexp-opt str-list)
                (mapconcat (lambda (s) (format "\\(%s\\)" s)) str-list "\\|"))))
    (mapcan (lambda (file)
              (howm-view-fake-grep-file reg file force-case-fold))
            (mapcan #'howm-view-files file-list))))

(defun howm-view-fake-grep-file (reg file force-case-fold)
  (let ((b (get-file-buffer file)))
    (if (and b howm-view-watch-modified-buffer)
        (with-current-buffer b
          (howm-view-fake-grep-current-buffer reg file force-case-fold))
      (with-temp-buffer
        (insert-file-contents file)
        (howm-view-fake-grep-current-buffer reg file force-case-fold)))))

(defun howm-view-fake-grep-current-buffer (reg file force-case-fold)
  (save-excursion
    (save-restriction
      (widen)
      (goto-char (point-max))
      (let* ((found nil)
             (case-fold-search (or force-case-fold (not (howm-capital-p reg)))))
        (while (re-search-backward reg nil t)
          (beginning-of-line)
          (setq found
                (cons (list file
                            (riffle-get-place)
                            (buffer-substring-no-properties (point)
                                                            (line-end-position)))
                      found)))
        found))))

(defvar howm-view-call-process-last-command nil)  ;; for debug
(defun howm-view-call-process (command args
                                       &optional
                                       expected-return-value stdin-string)
  (setq howm-view-call-process-last-command (cons command args))
  (with-temp-buffer
    (let ((r (howm-view-call-process-here command args stdin-string)))
      (when (and expected-return-value
                 (not (= expected-return-value r)))
        (error "Process returns %s instead of expected %s."
               r expected-return-value))
      (howm-view-buffer-lines))))

(defun howm-view-call-process-here (command args &optional stdin-string)
  (let* ((beg (point))
         (end (progn
                (insert (or stdin-string ""))
                (point)))
         (a `(,beg ,end ,command t (t nil) nil ,@args)))
    (apply #'call-process-region a)))

(defun howm-view-buffer-lines (&optional buf)
  (save-excursion
    (when buf
      (set-buffer buf))
    (split-string (buffer-substring (point-min) (point-max)) "\n")))

(defun howm-view-grep-parse-line (line)
  (if (string-match "^\\(\\([a-zA-Z]:/\\)?[^:]*\\):\\([0-9]*\\):\\(.*\\)$"
                    line)
      (let ((file (match-string 1 line))
            (line (string-to-number (match-string 3 line)))
            (content (match-string 4 line)))
        (list file line content))
    nil))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; sort

(defun howm-view-sort ()
  (interactive)
  (let* ((table howm-view-sort-methods)
         (command (completing-read "sort by: " table nil t)))
    (call-interactively (cdr (assoc command table)))))

(defun howm-view-sort-by-random (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'(lambda (dummy) (random)) #'< reverse-p))

(defun howm-view-sort-by-name (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'howm-view-item-basename #'string< reverse-p))

(defun howm-view-sort-by-name-match (&optional reverse-p regexp path-p)
  (interactive "P")
  (howm-view-sort-by-general-match (if path-p
                                       #'howm-view-item-filename
                                     #'howm-view-item-basename)
                                   reverse-p regexp))

(defvar howm-view-sort-by-date-ignore-regexp "^[a-zA-Z]")
(defun howm-view-sort-by-date (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'howm-view-item-basename #'string<
                          reverse-p)
  (howm-view-sort-by-general-match #'howm-view-item-basename t
                                   howm-view-sort-by-date-ignore-regexp))
(defun howm-view-sort-by-reverse-date (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-by-date (not reverse-p)))

(defun howm-view-sort-by-summary (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general #'howm-view-item-summary #'string< reverse-p))

(defun howm-view-sort-by-summary-match (&optional reverse-p regexp)
  (interactive "P")
  (howm-view-sort-by-general-match #'howm-view-item-summary
                                   reverse-p regexp))

(defun howm-view-sort-by-general-match (picker &optional reverse-p regexp)
  (let ((r (or regexp (read-from-minibuffer "Regexp: ")))
        (howm-view-s-b-g-m-matched nil)) ;; need unique name?? :-(
    (howm-view-sort-general (lambda (item)
                              (if (string-match r
                                                (funcall picker item))
                                  (progn
                                    (setq howm-view-s-b-g-m-matched t)
                                    1)
                                0))
                            #'>
                            reverse-p)
    howm-view-s-b-g-m-matched))

(defun howm-view-sort-by-reminder (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-general (lambda (item)
                            (let ((s (howm-view-item-summary item))
                                  (r (howm-reminder-regexp howm-reminder-types))
                                  (max-str (format-time-string
                                            howm-reminder-today-format
                                            (encode-time 59 59 23 31 12
                                                         howm-view-max-year))))
                              (if (string-match r s)
                                  (match-string-no-properties 0 s)
                                max-str)))
                          #'string< reverse-p))

;; nonsense
; (defun howm-view-sort-by-atime (&optional reverse-p)
;   (interactive "P")
;   (howm-view-sort-by-xtime 'a reverse-p))

;; needless
; (defun howm-view-sort-by-ctime (&optional reverse-p)
;   (interactive "P")
;   (howm-view-sort-by-xtime 'c reverse-p))

(defun howm-view-sort-by-mtime (&optional reverse-p)
  (interactive "P")
  (howm-view-sort-by-xtime 'm reverse-p))

(defun howm-view-sort-by-xtime (x reverse-p)
  (howm-view-sort-general (lambda (item)
                            (howm-view-xtime (howm-view-item-filename item)
                                             x))
;                           #'>
;                           reverse-p
                          #'howm-view-string>
                          reverse-p
                          ))

(defun howm-view-sort-general (evaluator comparer &optional reverse-p)
  (let* ((howm-view-s-g-comparer comparer) ;; need unique name?? :-(
         (cmp (if reverse-p
                  (lambda (a b) (funcall howm-view-s-g-comparer b a))
                howm-view-s-g-comparer))
         (sorted (howm-view-sort-items evaluator cmp
                                       (howm-view-item-list))))
    (howm-view-summary (howm-view-name) sorted)))

(defun howm-view-sort-items (evaluator comparer item-list)
  (let* ((orig (mapcar (lambda (item) (cons (funcall evaluator item) item))
                       item-list))
         (sorted (sort orig (lambda (x y)
                              (funcall comparer (car x) (car y))))))
    (mapcar #'cdr sorted)))

(defun howm-view-sort-reverse ()
  (interactive)
  (howm-view-summary (howm-view-name)
                    (reverse (howm-view-item-list))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Dired-X

(defvar howm-view-dired-buffer-name "*howm-dired*")
(defvar howm-view-dired-ls-command "ls")
(defvar howm-view-dired-ls-options '("-l"))

(defun dired-virtual (dir)
  (howm-inhibit-warning-in-compilation))

(defun howm-view-dired ()
  (interactive)
  (require 'dired-x)
  (when (not (member major-mode
                     '(howm-view-summary-mode howm-view-contents-mode)))
    (error "Invalid mode for this command."))
;;   ;; bug in emacs-21.3.50?
;;   (when (not (fboundp 'dired-insert-headerline))
;;     (defun dired-insert-headerline (dir);; also used by dired-insert-subdir
;;       ;; Insert DIR's headerline with no trailing slash, exactly like ls
;;       ;; would, and put cursor where dired-build-subdir-alist puts subdir
;;       ;; boundaries.
;;       (save-excursion (insert "  " (directory-file-name dir) ":\n"))))
  (let* ((files (mapcar (lambda (item)
                          (file-relative-name (howm-view-item-filename item)))
                        (howm-view-item-list)))
         (args (append howm-view-dired-ls-options files))
         (a `((howm-view-summary-mode . ,howm-view-summary-persistent)
              (howm-view-contents-mode . ,howm-view-contents-persistent)))
         (p (howm-view-persistent-p (cdr (assoc major-mode a)))))
    (if p
        (howm-view-restore-window-configuration)
      (howm-view-kill-buffer))
    (switch-to-buffer (get-buffer-create howm-view-dired-buffer-name))
    (setq buffer-read-only nil)
    (erase-buffer)
    (howm-view-call-process-here howm-view-dired-ls-command args)
    (set-buffer-modified-p nil)
    (dired-virtual default-directory)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; shell

(defvar howm-view-summary-shell-hist '("ls -l FILE" "FILE"))
(defvar howm-view-summary-shell-last-file "FILE")
(defun howm-view-summary-shell-command ()
  (interactive)
  (when (not (member major-mode
                     '(howm-view-summary-mode)))
    (error "Invalid mode for this command."))
  (let* ((n (howm-view-line-number))
         (item (nth (1- n) (howm-view-item-list)))
         (file (howm-abbreviate-file-name (howm-view-item-filename item)))
         (last-reg (regexp-quote howm-view-summary-shell-last-file)))
    (setq howm-view-summary-shell-hist
          (mapcar (lambda (h)
                    (replace-regexp-in-string last-reg file h t))
                  howm-view-summary-shell-hist))
    (setq howm-view-summary-shell-last-file file)
    (let* ((default (car howm-view-summary-shell-hist))
           (c (read-string "command: "
                           (cons default 0)
                           '(howm-view-summary-shell-hist . 1))))
      (shell-command c))
    (let ((item-list (remove-if (lambda (item)
                                  (not (file-exists-p
                                        (howm-view-item-filename item))))
                                (howm-view-item-list))))
      (setq *riffle-summary-check* nil) ;; dirty
      (howm-view-summary (howm-view-name) item-list)
      (goto-line n)
      (save-selected-window
        (let ((b (get-buffer "*Shell Command Output*")))
          (cond ((not (howm-buffer-empty-p b))
                 (switch-to-buffer-other-window b))
                ((eq item (riffle-summary-current-item))
                 nil)
                (t (progn
                     (setq *riffle-summary-check* t) ;; dirty
                     (howm-view-summary-check t))))))
      )))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; xemacs

(defun howm-try-require (feature)
  (and (locate-library (symbol-name feature))
       (require feature)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; external viewer

(defvar howm-view-external-viewer-assoc nil)
(defvar howm-view-use-mailcap
  (and (howm-try-require 'mailcap)
       (fboundp 'mailcap-parse-mailcaps)
       (fboundp 'mailcap-parse-mimetypes)))
(defvar howm-view-open-by-myself  ;; regexp list
  '("text/.*" "application/emacs-lisp"))
(when howm-view-use-mailcap
  (require 'mailcap)
  (mailcap-parse-mailcaps)
  (mailcap-parse-mimetypes))

(defun howm-view-external-viewer (file)
  (let ((dir-viewer (and (file-directory-p file) #'howm-view-directory-viewer))
        (viewer (cdr (assoc-if (lambda (reg) (string-match reg file))
                               howm-view-external-viewer-assoc))))
    (or dir-viewer viewer
        (and howm-view-use-mailcap
             (let* ((ext (if (string-match "\\.[^\\.]+$" file)
                             (match-string 0 file)
                           ""))
                    (type (mailcap-extension-to-mime ext))
                    (type-match (lambda (r) (string-match r type))))
               (cond ((null type) nil)
                     ((member-if type-match howm-view-open-by-myself) nil)
                     (t (mailcap-mime-info type))))))))

(defun howm-view-call-external-viewer (viewer file)
  (cond ((functionp viewer)
         (progn
           (find-file file)
           (funcall viewer)))
        ((stringp viewer)
         (start-process "howm-view-external-viewer" nil
                        shell-file-name
                        shell-command-switch
                        (format viewer file)))
        (t (error "Unknown viewer %s for type %s." viewer file))))

(defun howm-view-directory-viewer ()
  t) ;; NOP

(defadvice action-lock-find-file (around external-viewer (f u) activate)
  (let ((viewer (howm-view-external-viewer f)))
    (if viewer
        (howm-view-call-external-viewer viewer f)
      ad-do-it)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; provide

(provide 'howm-view)

;;; howm-view.el ends here
